@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2014 Jeffrey Stedfast
+ *  Copyright (C) 2000-2011 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -34,7 +34,6 @@
 #include "gmime-parse-utils.h"
 #include "gmime-stream-mem.h"
 #include "gmime-multipart.h"
-#include "gmime-common.h"
 #include "gmime-part.h"
 
 #if GLIB_MAJOR_VERSION > 2 || (GLIB_MAJOR_VERSION == 2 && GLIB_MINOR_VERSION >= 14)
@@ -93,9 +92,9 @@ static void parser_init (GMimeParser *parser, GMimeStream *stream);
 static void parser_close (GMimeParser *parser);
 
 static GMimeObject *parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type,
-						gboolean toplevel, int *found);
+						int *found);
 static GMimeObject *parser_construct_multipart (GMimeParser *parser, ContentType *content_type,
-						gboolean toplevel, int *found);
+						int *found);
 
 static GObjectClass *parent_class = NULL;
 
@@ -127,7 +126,7 @@ struct _GMimeParserPrivate {
 	gint64 offset;
 	
 	/* i/o buffers */
-	char realbuf[SCAN_HEAD + SCAN_BUF + 4];
+	char realbuf[SCAN_HEAD + SCAN_BUF + 1];
 	char *inbuf;
 	char *inptr;
 	char *inend;
@@ -180,8 +179,6 @@ struct _GMimeParserPrivate {
 	BoundaryStack *bounds;
 };
 
-static const char MBOX_BOUNDARY[6] = "From ";
-#define MBOX_BOUNDARY_LEN 5
 
 static void
 parser_push_boundary (GMimeParser *parser, const char *boundary)
@@ -196,10 +193,10 @@ parser_push_boundary (GMimeParser *parser, const char *boundary)
 	s->parent = priv->bounds;
 	priv->bounds = s;
 	
-	if (boundary == MBOX_BOUNDARY) {
-		s->boundary = g_strdup (boundary);
-		s->boundarylen = MBOX_BOUNDARY_LEN;
-		s->boundarylenfinal = MBOX_BOUNDARY_LEN;
+	if (!strcmp (boundary, "From ")) {
+		s->boundary = g_strdup ("From ");
+		s->boundarylen = 5;
+		s->boundarylenfinal = 5;
 	} else {
 		s->boundary = g_strdup_printf ("--%s--", boundary);
 		s->boundarylen = strlen (boundary) + 2;
@@ -514,8 +511,6 @@ g_mime_parser_get_persist_stream (GMimeParser *parser)
  * Note: This attribute only serves as a hint to the @parser. If the
  * underlying stream does not support seeking, then this attribute
  * will be ignored.
- *
- * By default, this feature is enabled if the underlying stream is seekable.
  **/
 void
 g_mime_parser_set_persist_stream (GMimeParser *parser, gboolean persist)
@@ -574,8 +569,6 @@ g_mime_parser_get_scan_from (GMimeParser *parser)
  * @scan_from: %TRUE to scan From-lines or %FALSE otherwise
  *
  * Sets whether or not @parser should scan mbox-style From-lines.
- *
- * By default, this feature is disabled.
  **/
 void
 g_mime_parser_set_scan_from (GMimeParser *parser, gboolean scan_from)
@@ -616,8 +609,6 @@ g_mime_parser_get_respect_content_length (GMimeParser *parser)
  *
  * Most notably useful when parsing broken Solaris mbox files (See
  * http://www.jwz.org/doc/content-length.html for details).
- *
- * By default, this feature is disabled.
  **/
 void
 g_mime_parser_set_respect_content_length (GMimeParser *parser, gboolean respect_content_length)
@@ -925,8 +916,10 @@ static void
 header_parse (GMimeParser *parser, HeaderRaw **tail)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	register char *inptr;
+	register char *inptr, *end;
 	HeaderRaw *header;
+	char *start;
+	size_t hlen;
 	
 	*priv->headerptr = '\0';
 	inptr = priv->headerbuf;
@@ -945,11 +938,28 @@ header_parse (GMimeParser *parser, HeaderRaw **tail)
 		return;
 	}
 	
+	hlen = inptr - priv->headerbuf;
+	
 	header = g_slice_new (HeaderRaw);
 	header->next = NULL;
 	
-	header->name = g_strndup (priv->headerbuf, (size_t) (inptr - priv->headerbuf));
-	header->value = g_mime_strdup_trim (inptr + 1);
+	header->name = g_strndup (priv->headerbuf, hlen);
+	
+	/* skip over leading lwsp */
+	inptr++;
+	while (is_lwsp (*inptr))
+		inptr++;
+	
+	/* cut trailing lwsp */
+	start = inptr;
+	end = inptr;
+	
+	while (*inptr) {
+		if (!is_lwsp (*inptr++))
+			end = inptr;
+	}
+	
+	header->value = g_strndup (start, (size_t) (end - start));
 	
 	header->offset = priv->header_offset;
 	
@@ -971,55 +981,6 @@ header_parse (GMimeParser *parser, HeaderRaw **tail)
 #endif
 }
 
-enum {
-	SUBJECT = 1 << 0,
-	FROM    = 1 << 1,
-	DATE    = 1 << 2,
-	TO      = 1 << 3,
-	CC      = 1 << 4
-};
-
-static gboolean
-has_message_headers (HeaderRaw *headers)
-{
-	unsigned int found = 0;
-	HeaderRaw *header;
-	
-	header = headers;
-	while (header != NULL) {
-		if (!g_ascii_strcasecmp (header->name, "Subject"))
-			found |= SUBJECT;
-		else if (!g_ascii_strcasecmp (header->name, "From"))
-			found |= FROM;
-		else if (!g_ascii_strcasecmp (header->name, "Date"))
-			found |= DATE;
-		else if (!g_ascii_strcasecmp (header->name, "To"))
-			found |= TO;
-		else if (!g_ascii_strcasecmp (header->name, "Cc"))
-			found |= CC;
-		
-		header = header->next;
-	}
-	
-	return found != 0;
-}
-
-static gboolean
-has_content_headers (HeaderRaw *headers)
-{
-	HeaderRaw *header;
-	
-	header = headers;
-	while (header != NULL) {
-		if (!g_ascii_strcasecmp (header->name, "Content-Type"))
-			return TRUE;
-		
-		header = header->next;
-	}
-	
-	return FALSE;
-}
-
 static int
 parser_step_headers (GMimeParser *parser)
 {
@@ -1101,26 +1062,16 @@ parser_step_headers (GMimeParser *parser)
 						goto next_message;
 					
 					if (priv->headers != NULL) {
-						if (priv->state == GMIME_PARSER_STATE_MESSAGE_HEADERS) {
-							if (has_message_headers (priv->headers)) {
-								/* probably the start of the content,
-								 * a broken mailer didn't terminate the
-								 * headers with an empty line. *sigh* */
-								goto content_start;
-							}
-						} else if (has_content_headers (priv->headers)) {
-							/* probably the start of the content,
-							 * a broken mailer didn't terminate the
-							 * headers with an empty line. *sigh* */
-							goto content_start;
-						}
-					} else if (priv->state == GMIME_PARSER_STATE_MESSAGE_HEADERS) {
-						/* Be a little more strict when scanning toplevel message
-						 * headers, but remain lenient with From-lines. */
-						if ((inptr - start) != 4 || strncmp (start, "From ", 5) != 0) {
-							priv->state = GMIME_PARSER_STATE_ERROR;
-							return -1;
-						}
+						/* probably the start of the content,
+						 * a broken mailer didn't terminate the
+						 * headers with an empty line. *sigh* */
+						goto content_start;
+					}
+					
+					if (priv->state == GMIME_PARSER_STATE_MESSAGE_HEADERS) {
+						/* Be a little more strict when scanning toplevel message headers. */
+						priv->state = GMIME_PARSER_STATE_ERROR;
+						return -1;
 					}
 				}
 			}
@@ -1174,7 +1125,6 @@ parser_step_headers (GMimeParser *parser)
 	
 	inptr = priv->inptr;
 	inend = priv->inend;
-	start = inptr;
 	
 	len = (size_t) (inend - inptr);
 	header_append (priv, inptr, len);
@@ -1240,7 +1190,7 @@ content_type_is_type (ContentType *content_type, const char *type, const char *s
 }
 
 static ContentType *
-parser_content_type (GMimeParser *parser, GMimeContentType *parent)
+parser_content_type (GMimeParser *parser)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	ContentType *content_type;
@@ -1250,13 +1200,8 @@ parser_content_type (GMimeParser *parser, GMimeContentType *parent)
 	
 	if (!(value = header_raw_find (priv->headers, "Content-Type", NULL)) ||
 	    !g_mime_parse_content_type (&value, &content_type->type, &content_type->subtype)) {
-		if (parent != NULL && g_mime_content_type_is_type (parent, "multipart", "digest")) {
-			content_type->type = g_strdup ("message");
-			content_type->subtype = g_strdup ("rfc822");
-		} else {
-			content_type->type = g_strdup ("text");
-			content_type->subtype = g_strdup ("plain");
-		}
+		content_type->type = g_strdup ("text");
+		content_type->subtype = g_strdup ("plain");
 	}
 	
 	content_type->exists = value != NULL;
@@ -1364,33 +1309,6 @@ enum {
                          ((scan_from && len >= 5 && !strncmp (start, "From ", 5)) ||  \
 			  (len >= 2 && (start[0] == '-' && start[1] == '-')))
 
-static gboolean
-is_boundary (const char *text, size_t len, const char *boundary, size_t boundary_len)
-{
-	const char *inptr = text + boundary_len;
-	const char *inend = text + len;
-	
-	if (boundary_len > len)
-		return FALSE;
-	
-	/* make sure that the text matches the boundary */
-	if (strncmp (text, boundary, boundary_len) != 0)
-		return FALSE;
-	
-	if (!strncmp (text, "From ", 5))
-		return TRUE;
-	
-	/* the boundary may be optionally followed by linear whitespace */
-	while (inptr < inend) {
-		if (!is_lwsp (*inptr))
-			return FALSE;
-		
-		inptr++;
-	}
-	
-	return TRUE;
-}
-
 static int
 check_boundary (struct _GMimeParserPrivate *priv, const char *start, size_t len)
 {
@@ -1406,14 +1324,19 @@ check_boundary (struct _GMimeParserPrivate *priv, const char *start, size_t len)
 		
 		s = priv->bounds;
 		while (s) {
+			/* we use >= here because From lines are > 5 chars */
 			if (offset >= s->content_end &&
-			    is_boundary (start, len, s->boundary, s->boundarylenfinal)) {
+			    len >= s->boundarylenfinal &&
+			    !strncmp (s->boundary, start,
+				      s->boundarylenfinal)) {
 				d(printf ("found %s\n", s->content_end != -1 && offset >= s->content_end ?
 					  "end of content" : "end boundary"));
 				return FOUND_END_BOUNDARY;
 			}
 			
-			if (is_boundary (start, len, s->boundary, s->boundarylen)) {
+			if (len == s->boundarylen &&
+			    !strncmp (s->boundary, start,
+				      s->boundarylen)) {
 				d(printf ("found boundary\n"));
 				return FOUND_BOUNDARY;
 			}
@@ -1427,22 +1350,6 @@ check_boundary (struct _GMimeParserPrivate *priv, const char *start, size_t len)
 	return FOUND_NOTHING;
 }
 
-static gboolean
-found_immediate_boundary (struct _GMimeParserPrivate *priv, gboolean end)
-{
-	BoundaryStack *s = priv->bounds;
-	size_t boundary_len = end ? s->boundarylenfinal : s->boundarylen;
-	register char *inptr = priv->inptr;
-	char *inend = priv->inend;
-	
-	/* Note: see optimization comment [1] */
-	*inend = '\n';
-	while (*inptr != '\n')
-		inptr++;
-	
-	return is_boundary (priv->inptr, inptr - priv->inptr, s->boundary, boundary_len);
-}
-
 /* Optimization Notes:
  *
  * 1. By making the priv->realbuf char array 1 extra char longer, we
@@ -1461,13 +1368,11 @@ static int
 parser_scan_content (GMimeParser *parser, GByteArray *content, guint *crlf)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	char *aligned, *start, *inend;
 	register char *inptr;
-	register int *dword;
+	char *start, *inend;
 	size_t nleft, len;
 	size_t atleast;
 	int found = 0;
-	int mask;
 	
 	d(printf ("scan-content\n"));
 	
@@ -1501,26 +1406,11 @@ parser_scan_content (GMimeParser *parser, GByteArray *content, guint *crlf)
 		priv->midline = FALSE;
 		
 		while (inptr < inend) {
-			aligned = (char *) (((long) (inptr + 3)) & ~3);
 			start = inptr;
-			
 			/* Note: see optimization comment [1] */
-			while (inptr < aligned && *inptr != '\n')
+			while (*inptr != '\n')
 				inptr++;
 			
-			if (inptr == aligned) {
-				dword = (int *) inptr;
-				
-				do {
-					mask = *dword++ ^ 0x0A0A0A0A;
-					mask = ((mask - 0x01010101) & (~mask & 0x80808080));
-				} while (mask == 0);
-				
-				inptr = (char *) (dword - 1);
-				while (*inptr != '\n')
-					inptr++;
-			}
-			
 			len = (size_t) (inptr - start);
 			
 			if (inptr < inend) {
@@ -1624,40 +1514,6 @@ parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *fou
 	
 	g_assert (priv->state == GMIME_PARSER_STATE_CONTENT);
 	
-	if (priv->bounds != NULL) {
-		/* Check for the possibility of an empty message/rfc822 part. */
-		register char *inptr;
-		size_t atleast;
-		char *inend;
-		
-		/* figure out minimum amount of data we need */
-		atleast = MAX (SCAN_HEAD, MAX_BOUNDARY_LEN (priv->bounds));
-		
-		if (parser_fill (parser, atleast) <= 0) {
-			*found = FOUND_EOS;
-			return;
-		}
-		
-		inptr = priv->inptr;
-		inend = priv->inend;
-		/* Note: see optimization comment [1] */
-		*inend = '\n';
-		
-		while (*inptr != '\n')
-			inptr++;
-		
-		*found = check_boundary (priv, priv->inptr, inptr - priv->inptr);
-		switch (*found) {
-		case FOUND_END_BOUNDARY:
-			/* ignore "From " boundaries, boken mailers tend to include these lines... */
-			if (strncmp (priv->inptr, "From ", 5) != 0)
-				return;
-			break;
-		case FOUND_BOUNDARY:
-			return;
-		}
-	}
-	
 	/* get the headers */
 	priv->state = GMIME_PARSER_STATE_HEADERS;
 	if (parser_step (parser) == GMIME_PARSER_STATE_ERROR) {
@@ -1670,16 +1526,15 @@ parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *fou
 	message = g_mime_message_new (FALSE);
 	header = priv->headers;
 	while (header) {
-		if (g_ascii_strncasecmp (header->name, "Content-", 8) != 0)
-			g_mime_object_append_header ((GMimeObject *) message, header->name, header->value);
+		g_mime_object_append_header ((GMimeObject *) message, header->name, header->value);
 		header = header->next;
 	}
 	
-	content_type = parser_content_type (parser, NULL);
+	content_type = parser_content_type (parser);
 	if (content_type_is_type (content_type, "multipart", "*"))
-		object = parser_construct_multipart (parser, content_type, TRUE, found);
+		object = parser_construct_multipart (parser, content_type, found);
 	else
-		object = parser_construct_leaf_part (parser, content_type, TRUE, found);
+		object = parser_construct_leaf_part (parser, content_type, found);
 	
 	content_type_destroy (content_type);
 	message->mime_part = object;
@@ -1693,7 +1548,7 @@ parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *fou
 }
 
 static GMimeObject *
-parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type, gboolean toplevel, int *found)
+parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type, int *found)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	GMimeObject *object;
@@ -1707,15 +1562,14 @@ parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type, gboo
 	if (!content_type->exists) {
 		GMimeContentType *mime_type;
 		
-		mime_type = g_mime_content_type_new (content_type->type, content_type->subtype);
+		mime_type = g_mime_content_type_new ("text", "plain");
 		_g_mime_object_set_content_type (object, mime_type);
 		g_object_unref (mime_type);
 	}
 	
 	header = priv->headers;
 	while (header) {
-		if (!toplevel || !g_ascii_strncasecmp (header->name, "Content-", 8))
-			g_mime_object_append_header (object, header->name, header->value);
+		g_mime_object_append_header (object, header->name, header->value);
 		header = header->next;
 	}
 	
@@ -1805,6 +1659,16 @@ parser_scan_multipart_face (GMimeParser *parser, GMimeMultipart *multipart, gboo
 #define parser_scan_multipart_preface(parser, multipart) parser_scan_multipart_face (parser, multipart, TRUE)
 #define parser_scan_multipart_postface(parser, multipart) parser_scan_multipart_face (parser, multipart, FALSE)
 
+static gboolean
+found_immediate_boundary (struct _GMimeParserPrivate *priv, gboolean end)
+{
+	BoundaryStack *s = priv->bounds;
+	size_t len = end ? s->boundarylenfinal : s->boundarylen;
+	
+	return !strncmp (priv->inptr, s->boundary, len)
+		&& (priv->inptr[len] == '\n' || priv->inptr[len] == '\r');
+}
+
 static int
 parser_scan_multipart_subparts (GMimeParser *parser, GMimeMultipart *multipart)
 {
@@ -1832,11 +1696,11 @@ parser_scan_multipart_subparts (GMimeParser *parser, GMimeMultipart *multipart)
 			break;
 		}
 		
-		content_type = parser_content_type (parser, ((GMimeObject *) multipart)->content_type);
+		content_type = parser_content_type (parser);
 		if (content_type_is_type (content_type, "multipart", "*"))
-			subpart = parser_construct_multipart (parser, content_type, FALSE, &found);
+			subpart = parser_construct_multipart (parser, content_type, &found);
 		else
-			subpart = parser_construct_leaf_part (parser, content_type, FALSE, &found);
+			subpart = parser_construct_leaf_part (parser, content_type, &found);
 		
 		g_mime_multipart_add (multipart, subpart);
 		content_type_destroy (content_type);
@@ -1847,7 +1711,7 @@ parser_scan_multipart_subparts (GMimeParser *parser, GMimeMultipart *multipart)
 }
 
 static GMimeObject *
-parser_construct_multipart (GMimeParser *parser, ContentType *content_type, gboolean toplevel, int *found)
+parser_construct_multipart (GMimeParser *parser, ContentType *content_type, int *found)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	GMimeMultipart *multipart;
@@ -1862,8 +1726,7 @@ parser_construct_multipart (GMimeParser *parser, ContentType *content_type, gboo
 	
 	header = priv->headers;
 	while (header) {
-		if (!toplevel || !g_ascii_strncasecmp (header->name, "Content-", 8))
-			g_mime_object_append_header (object, header->name, header->value);
+		g_mime_object_append_header (object, header->name, header->value);
 		header = header->next;
 	}
 	
@@ -1931,11 +1794,11 @@ parser_construct_part (GMimeParser *parser)
 			return NULL;
 	}
 	
-	content_type = parser_content_type (parser, NULL);
+	content_type = parser_content_type (parser);
 	if (content_type_is_type (content_type, "multipart", "*"))
-		object = parser_construct_multipart (parser, content_type, FALSE, &found);
+		object = parser_construct_multipart (parser, content_type, &found);
 	else
-		object = parser_construct_leaf_part (parser, content_type, FALSE, &found);
+		object = parser_construct_leaf_part (parser, content_type, &found);
 	
 	content_type_destroy (content_type);
 	
@@ -1949,8 +1812,7 @@ parser_construct_part (GMimeParser *parser)
  *
  * Constructs a MIME part from @parser.
  *
- * Returns: (transfer full): a MIME part based on @parser or %NULL on
- * fail.
+ * Returns: a MIME part based on @parser or %NULL on fail.
  **/
 GMimeObject *
 g_mime_parser_construct_part (GMimeParser *parser)
@@ -1995,22 +1857,21 @@ parser_construct_message (GMimeParser *parser)
 				content_length = ULONG_MAX;
 		}
 		
-		if (g_ascii_strncasecmp (header->name, "Content-", 8) != 0)
-			g_mime_object_append_header ((GMimeObject *) message, header->name, header->value);
+		g_mime_object_append_header ((GMimeObject *) message, header->name, header->value);
 		header = header->next;
 	}
 	
 	if (priv->scan_from) {
-		parser_push_boundary (parser, MBOX_BOUNDARY);
+		parser_push_boundary (parser, "From ");
 		if (priv->respect_content_length && content_length < ULONG_MAX)
 			priv->bounds->content_end = parser_offset (priv, NULL) + content_length;
 	}
 	
-	content_type = parser_content_type (parser, NULL);
+	content_type = parser_content_type (parser);
 	if (content_type_is_type (content_type, "multipart", "*"))
-		object = parser_construct_multipart (parser, content_type, TRUE, &found);
+		object = parser_construct_multipart (parser, content_type, &found);
 	else
-		object = parser_construct_leaf_part (parser, content_type, TRUE, &found);
+		object = parser_construct_leaf_part (parser, content_type, &found);
 	
 	content_type_destroy (content_type);
 	message->mime_part = object;
@@ -2034,7 +1895,7 @@ parser_construct_message (GMimeParser *parser)
  *
  * Constructs a MIME message from @parser.
  *
- * Returns: (transfer full): a MIME message or %NULL on fail.
+ * Returns: a MIME message or %NULL on fail.
  **/
 GMimeMessage *
 g_mime_parser_construct_message (GMimeParser *parser)
