@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2009 Jeffrey Stedfast
+ *  Copyright (C) 2000-2011 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -32,6 +32,7 @@
 #include "gmime-table-private.h"
 #include "gmime-parse-utils.h"
 #include "gmime-iconv-utils.h"
+#include "gmime-events.h"
 #include "gmime-utils.h"
 #include "list.h"
 
@@ -91,131 +92,6 @@ enum {
 };
 
 
-typedef void (* EventCallback) (gpointer sender, gpointer args);
-
-typedef struct _EventListener {
-	struct _EventListener *next;
-	struct _EventListener *prev;
-	EventCallback callback;
-	gpointer user_data;
-	int blocked;
-} EventListener;
-
-static EventListener *
-event_listener_new (EventCallback callback, gpointer user_data)
-{
-	EventListener *listener;
-	
-	listener = g_slice_new (EventListener);
-	listener->user_data = user_data;
-	listener->callback = callback;
-	listener->prev = NULL;
-	listener->next = NULL;
-	listener->blocked = 0;
-	
-	return listener;
-}
-
-static void
-event_listener_free (EventListener *listener)
-{
-	g_slice_free (EventListener, listener);
-}
-
-
-static List *
-event_list_new (void)
-{
-	List *list;
-	
-	list = g_slice_new (List);
-	list_init (list);
-	
-	return list;
-}
-
-static void
-event_list_free (List *list)
-{
-	EventListener *node, *next;
-	
-	node = (EventListener *) list->head;
-	while (node->next) {
-		next = node->next;
-		event_listener_free (node);
-		node = next;
-	}
-	
-	g_slice_free (List, list);
-}
-
-static EventListener *
-event_list_find_listener (List *list, EventCallback callback, gpointer user_data)
-{
-	EventListener *node;
-	
-	node = (EventListener *) list->head;
-	while (node->next) {
-		if (node->callback == callback && node->user_data == user_data)
-			return node;
-		node = node->next;
-	}
-	
-	return NULL;
-}
-
-static void
-event_list_block (List *list, EventCallback callback, gpointer user_data)
-{
-	EventListener *listener;
-	
-	if ((listener = event_list_find_listener (list, callback, user_data)))
-		listener->blocked++;
-}
-
-static void
-event_list_unblock (List *list, EventCallback callback, gpointer user_data)
-{
-	EventListener *listener;
-	
-	if ((listener = event_list_find_listener (list, callback, user_data)))
-		listener->blocked--;
-}
-
-static void
-event_list_add (List *list, EventCallback callback, gpointer user_data)
-{
-	EventListener *listener;
-	
-	listener = event_listener_new (callback, user_data);
-	list_append (list, (ListNode *) listener);
-}
-
-static void
-event_list_remove (List *list, EventCallback callback, gpointer user_data)
-{
-	EventListener *listener;
-	
-	if ((listener = event_list_find_listener (list, callback, user_data))) {
-		list_unlink ((ListNode *) listener);
-		event_listener_free (listener);
-	}
-}
-
-static void
-event_list_emit (List *list, gpointer sender)
-{
-	EventListener *node;
-	
-	node = (EventListener *) list->head;
-	while (node->next) {
-		if (node->blocked <= 0)
-			node->callback (sender, node->user_data);
-		node = node->next;
-	}
-}
-
-
 static void internet_address_class_init (InternetAddressClass *klass);
 static void internet_address_init (InternetAddress *ia, InternetAddressClass *klass);
 static void internet_address_finalize (GObject *object);
@@ -265,7 +141,7 @@ internet_address_class_init (InternetAddressClass *klass)
 static void
 internet_address_init (InternetAddress *ia, InternetAddressClass *klass)
 {
-	ia->priv = event_list_new ();
+	ia->priv = g_mime_event_new ((GObject *) ia);
 	ia->name = NULL;
 }
 
@@ -274,7 +150,7 @@ internet_address_finalize (GObject *object)
 {
 	InternetAddress *ia = (InternetAddress *) object;
 	
-	event_list_free (ia->priv);
+	g_mime_event_destroy (ia->priv);
 	g_free (ia->name);
 	
 	G_OBJECT_CLASS (parent_class)->finalize (object);
@@ -286,15 +162,9 @@ _internet_address_set_name (InternetAddress *ia, const char *name)
 {
 	char *buf;
 	
-	if (name) {
-		buf = g_mime_utils_header_decode_phrase (name);
-		g_mime_utils_unquote_string (buf);
-		g_free (ia->name);
-		ia->name = buf;
-	} else {
-		g_free (ia->name);
-		ia->name = NULL;
-	}
+	buf = g_strdup (name);
+	g_free (ia->name);
+	ia->name = buf;
 }
 
 /**
@@ -311,7 +181,7 @@ internet_address_set_name (InternetAddress *ia, const char *name)
 	
 	_internet_address_set_name (ia, name);
 	
-	event_list_emit (ia->priv, ia);
+	g_mime_event_emit (ia->priv, NULL);
 }
 
 
@@ -446,8 +316,7 @@ internet_address_mailbox_new (const char *name, const char *addr)
 	mailbox = g_object_newv (INTERNET_ADDRESS_TYPE_MAILBOX, 0, NULL);
 	mailbox->addr = g_strdup (addr);
 	
-	if (name != NULL)
-		_internet_address_set_name ((InternetAddress *) mailbox, name);
+	_internet_address_set_name ((InternetAddress *) mailbox, name);
 	
 	return (InternetAddress *) mailbox;
 }
@@ -471,7 +340,7 @@ internet_address_mailbox_set_addr (InternetAddressMailbox *mailbox, const char *
 	g_free (mailbox->addr);
 	mailbox->addr = g_strdup (addr);
 	
-	event_list_emit (((InternetAddress *) mailbox)->priv, mailbox);
+	g_mime_event_emit (((InternetAddress *) mailbox)->priv, NULL);
 }
 
 
@@ -541,9 +410,9 @@ internet_address_group_class_init (InternetAddressGroupClass *klass)
 }
 
 static void
-members_changed (InternetAddressList *members, InternetAddress *group)
+members_changed (InternetAddressList *members, gpointer args, InternetAddress *group)
 {
-	event_list_emit (((InternetAddress *) group)->priv, group);
+	g_mime_event_emit (((InternetAddress *) group)->priv, NULL);
 }
 
 static void
@@ -551,7 +420,7 @@ internet_address_group_init (InternetAddressGroup *group, InternetAddressGroupCl
 {
 	group->members = internet_address_list_new ();
 	
-	event_list_add (group->members->priv, (EventCallback) members_changed, group);
+	g_mime_event_add (group->members->priv, (GMimeEventCallback) members_changed, group);
 }
 
 static void
@@ -559,7 +428,7 @@ internet_address_group_finalize (GObject *object)
 {
 	InternetAddressGroup *group = (InternetAddressGroup *) object;
 	
-	event_list_remove (group->members->priv, (EventCallback) members_changed, group);
+	g_mime_event_remove (group->members->priv, (GMimeEventCallback) members_changed, group);
 	
 	g_object_unref (group->members);
 	
@@ -605,18 +474,18 @@ internet_address_group_set_members (InternetAddressGroup *group, InternetAddress
 		return;
 	
 	if (group->members) {
-		event_list_remove (group->members->priv, (EventCallback) members_changed, group);
+		g_mime_event_remove (group->members->priv, (GMimeEventCallback) members_changed, group);
 		g_object_unref (group->members);
 	}
 	
 	if (members) {
-		event_list_add (members->priv, (EventCallback) members_changed, group);
+		g_mime_event_add (members->priv, (GMimeEventCallback) members_changed, group);
 		g_object_ref (members);
 	}
 	
 	group->members = members;
 	
-	event_list_emit (((InternetAddress *) group)->priv, group);
+	g_mime_event_emit (((InternetAddress *) group)->priv, NULL);
 }
 
 
@@ -705,14 +574,14 @@ internet_address_list_class_init (InternetAddressListClass *klass)
 static void
 internet_address_list_init (InternetAddressList *list, InternetAddressListClass *klass)
 {
-	list->priv = event_list_new ();
+	list->priv = g_mime_event_new ((GObject *) list);
 	list->array = g_ptr_array_new ();
 }
 
 static void
-address_changed (InternetAddress *ia, InternetAddressList *list)
+address_changed (InternetAddress *ia, gpointer args, InternetAddressList *list)
 {
-	event_list_emit (list->priv, list);
+	g_mime_event_emit (list->priv, NULL);
 }
 
 static void
@@ -724,11 +593,11 @@ internet_address_list_finalize (GObject *object)
 	
 	for (i = 0; i < list->array->len; i++) {
 		ia = (InternetAddress *) list->array->pdata[i];
-		event_list_remove (ia->priv, (EventCallback) address_changed, list);
+		g_mime_event_remove (ia->priv, (GMimeEventCallback) address_changed, list);
 		g_object_unref (ia);
 	}
 	
-	event_list_free (list->priv);
+	g_mime_event_destroy (list->priv);
 	
 	g_ptr_array_free (list->array, TRUE);
 	
@@ -783,13 +652,13 @@ internet_address_list_clear (InternetAddressList *list)
 	
 	for (i = 0; i < list->array->len; i++) {
 		ia = (InternetAddress *) list->array->pdata[i];
-		event_list_remove (ia->priv, (EventCallback) address_changed, list);
+		g_mime_event_remove (ia->priv, (GMimeEventCallback) address_changed, list);
 		g_object_unref (ia);
 	}
 	
 	g_ptr_array_set_size (list->array, 0);
 	
-	event_list_emit (list->priv, list);
+	g_mime_event_emit (list->priv, NULL);
 }
 
 
@@ -798,7 +667,7 @@ _internet_address_list_add (InternetAddressList *list, InternetAddress *ia)
 {
 	int index;
 	
-	event_list_add (ia->priv, (EventCallback) address_changed, list);
+	g_mime_event_add (ia->priv, (GMimeEventCallback) address_changed, list);
 	
 	index = list->array->len;
 	g_ptr_array_add (list->array, ia);
@@ -827,7 +696,7 @@ internet_address_list_add (InternetAddressList *list, InternetAddress *ia)
 	index = _internet_address_list_add (list, ia);
 	g_object_ref (ia);
 	
-	event_list_emit (list->priv, list);
+	g_mime_event_emit (list->priv, NULL);
 	
 	return index;
 }
@@ -863,12 +732,12 @@ internet_address_list_prepend (InternetAddressList *list, InternetAddressList *p
 	
 	for (i = 0; i < prepend->array->len; i++) {
 		ia = (InternetAddress *) prepend->array->pdata[i];
-		event_list_add (ia->priv, (EventCallback) address_changed, list);
+		g_mime_event_add (ia->priv, (GMimeEventCallback) address_changed, list);
 		list->array->pdata[i] = ia;
 		g_object_ref (ia);
 	}
 	
-	event_list_emit (list->priv, list);
+	g_mime_event_emit (list->priv, NULL);
 }
 
 
@@ -893,12 +762,12 @@ internet_address_list_append (InternetAddressList *list, InternetAddressList *ap
 	
 	for (i = 0; i < append->array->len; i++) {
 		ia = (InternetAddress *) append->array->pdata[i];
-		event_list_add (ia->priv, (EventCallback) address_changed, list);
+		g_mime_event_add (ia->priv, (GMimeEventCallback) address_changed, list);
 		list->array->pdata[len + i] = ia;
 		g_object_ref (ia);
 	}
 	
-	event_list_emit (list->priv, list);
+	g_mime_event_emit (list->priv, NULL);
 }
 
 
@@ -921,7 +790,7 @@ internet_address_list_insert (InternetAddressList *list, int index, InternetAddr
 	g_return_if_fail (IS_INTERNET_ADDRESS (ia));
 	g_return_if_fail (index >= 0);
 	
-	event_list_add (ia->priv, (EventCallback) address_changed, list);
+	g_mime_event_add (ia->priv, (GMimeEventCallback) address_changed, list);
 	g_object_ref (ia);
 	
 	if ((guint) index < list->array->len) {
@@ -938,7 +807,7 @@ internet_address_list_insert (InternetAddressList *list, int index, InternetAddr
 		g_ptr_array_add (list->array, ia);
 	}
 	
-	event_list_emit (list->priv, list);
+	g_mime_event_emit (list->priv, NULL);
 }
 
 
@@ -992,12 +861,12 @@ internet_address_list_remove_at (InternetAddressList *list, int index)
 		return FALSE;
 	
 	ia = list->array->pdata[index];
-	event_list_remove (ia->priv, (EventCallback) address_changed, list);
+	g_mime_event_remove (ia->priv, (GMimeEventCallback) address_changed, list);
 	g_object_unref (ia);
 	
 	g_ptr_array_remove_index (list->array, index);
 	
-	event_list_emit (list->priv, list);
+	g_mime_event_emit (list->priv, NULL);
 	
 	return TRUE;
 }
@@ -1101,14 +970,14 @@ internet_address_list_set_address (InternetAddressList *list, int index, Interne
 	if ((old = list->array->pdata[index]) == ia)
 		return;
 	
-	event_list_remove (old->priv, (EventCallback) address_changed, list);
+	g_mime_event_remove (old->priv, (GMimeEventCallback) address_changed, list);
 	g_object_unref (old);
 	
-	event_list_add (ia->priv, (EventCallback) address_changed, list);
+	g_mime_event_add (ia->priv, (GMimeEventCallback) address_changed, list);
 	list->array->pdata[index] = ia;
 	g_object_ref (ia);
 	
-	event_list_emit (list->priv, list);
+	g_mime_event_emit (list->priv, NULL);
 }
 
 
@@ -1219,7 +1088,7 @@ mailbox_to_string (InternetAddress *ia, guint32 flags, size_t *linelen, GString
 				*linelen += len;
 			}
 		} else {
-			/* we can safly fit the name on this line */
+			/* we can safely fit the name on this line */
 			g_string_append_len (string, name, len);
 			*linelen += len;
 		}
@@ -1276,18 +1145,21 @@ group_to_string (InternetAddress *ia, guint32 flags, size_t *linelen, GString *s
 	InternetAddressGroup *group = (InternetAddressGroup *) ia;
 	gboolean encode = flags & INTERNET_ADDRESS_ENCODE;
 	gboolean fold = flags & INTERNET_ADDRESS_FOLD;
-	char *name;
-	size_t len;
+	char *name = NULL;
+	size_t len = 0;
 	
-	name = encoded_name (ia->name, encode);
-	len = strlen (name);
-	
-	if (fold && *linelen > 1 && (*linelen + len + 1) > GMIME_FOLD_LEN) {
-		linewrap (string);
-		*linelen = 1;
+	if (ia->name != NULL) {
+		name = encoded_name (ia->name, encode);
+		len = strlen (name);
+		
+		if (fold && *linelen > 1 && (*linelen + len + 1) > GMIME_FOLD_LEN) {
+			linewrap (string);
+			*linelen = 1;
+		}
+		
+		g_string_append_len (string, name, len);
 	}
 	
-	g_string_append_len (string, name, len);
 	g_string_append_len (string, ": ", 2);
 	*linelen += len + 2;
 	g_free (name);
@@ -1352,82 +1224,70 @@ internet_address_list_writer (InternetAddressList *list, GString *str)
 	_internet_address_list_to_string (list, flags, &linelen, str);
 }
 
+static void
+_internet_address_decode_name (InternetAddress *ia, GString *name)
+{
+	char *value, *buf = NULL;
+	char *phrase;
+	
+	if (!g_utf8_validate (name->str, name->len, NULL)) {
+		/* A (broken) mailer has sent us raw 8bit/multibyte text data... */
+		buf = g_mime_utils_decode_8bit (name->str, name->len);
+		phrase = buf;
+	} else {
+		phrase = name->str;
+	}
+	
+	/* decode the phrase */
+	g_mime_utils_unquote_string (phrase);
+	value = g_mime_utils_header_decode_phrase (phrase);
+	g_free (ia->name);
+	ia->name = value;
+	g_free (buf);
+}
+
+static InternetAddress *decode_address (const char **in);
+
+static void
+skip_lwsp (const char **in)
+{
+	register const char *inptr = *in;
+	
+	while (*inptr && is_lwsp (*inptr))
+		inptr++;
+	
+	*in = inptr;
+}
+
 static InternetAddress *
-decode_mailbox (const char **in)
+decode_addrspec (const char **in)
 {
 	InternetAddress *mailbox = NULL;
-	const char *inptr, *word;
-	gboolean bracket = FALSE;
-	GString *name = NULL;
+	const char *start, *inptr, *word;
+	gboolean got_local = FALSE;
 	GString *addr;
-	size_t n = 0;
+	size_t len;
 	
 	addr = g_string_new ("");
-	
-	decode_lwsp (in);
 	inptr = *in;
 	
-	if ((word = decode_word (&inptr)))
-		n = inptr - word;
-	
 	decode_lwsp (&inptr);
-	if (*inptr && !strchr (",.@", *inptr)) {
-		gboolean retried = FALSE;
-		
-		/* this mailbox has a name part, so get the name */
-		name = g_string_new ("");
-		while (word) {
-			g_string_append_len (name, word, n);
-			retried = FALSE;
-		retry:
-			if ((word = decode_word (&inptr))) {
-				g_string_append_c (name, ' ');
-				n = inptr - word;
-			}
-		}
-		
+	
+	/* some spam bots set their addresses to stuff like: ).ORHH@em.ca */
+	while (*inptr && !(*inptr == '"' || is_atom (*inptr)))
+		inptr++;
+	
+	start = inptr;
+	
+	/* extract the first word of the local-part */
+	if ((word = decode_word (&inptr))) {
+		g_string_append_len (addr, word, (size_t) (inptr - word));
 		decode_lwsp (&inptr);
-		if (*inptr == '<') {
-			inptr++;
-			bracket = TRUE;
-			if ((word = decode_word (&inptr)))
-				n = inptr - word;
-		} else if (!retried && *inptr) {
-			w(g_warning ("Unexpected char '%c' in mailbox: %s: attempting recovery.",
-				     *inptr, *in));
-			/* chew up this bad char and then attempt 1 more pass at parsing */
-			g_string_append_c (name, *inptr++);
-			retried = TRUE;
-			goto retry;
-		} else {
-			g_string_free (name, TRUE);
-			g_string_free (addr, TRUE);
-			*in = inptr;
-			
-			return NULL;
-		}
+		got_local = TRUE;
 	}
 	
-	if (word) {
-		g_string_append_len (addr, word, n);
-	} else {
-		w(g_warning ("No local part for email address: %s", *in));
-		if (name)
-			g_string_free (name, TRUE);
-		g_string_free (addr, TRUE);
-		
-		/* comma will be eaten by our caller */
-		if (*inptr && *inptr != ',')
-			*in = inptr + 1;
-		else
-			*in = inptr;
-		
-		return NULL;
-	}
-	
-	/* get the rest of the local-part */
-	decode_lwsp (&inptr);
-	while (*inptr == '.' && word) {
+	/* extract the rest of the local-part */
+	while (word && *inptr == '.') {
 		/* Note: According to the spec, only a single '.' is
 		 * allowed between word tokens in the local-part of an
 		 * addr-spec token, but according to Evolution bug
@@ -1445,145 +1305,237 @@ decode_mailbox (const char **in)
 		decode_lwsp (&inptr);
 	}
 	
-	/* we should be at the '@' now... */
 	if (*inptr == '@') {
-		size_t len = addr->len;
+		len = addr->len;
 		
 		g_string_append_c (addr, '@');
 		inptr++;
 		
-		if (!decode_domain (&inptr, addr))
+		if (!decode_domain (&inptr, addr)) {
+			/* drop the @domain and continue as if it weren't there */
+			w(g_warning ("Missing domain in addr-spec: %.*s",
+				     inptr - start, start));
 			g_string_truncate (addr, len);
-	} else {
-		w(g_warning ("No domain in email address: %s", *in));
-	}
-	
-	if (bracket) {
-		decode_lwsp (&inptr);
-		if (*inptr == '>')
-			inptr++;
-		else
-			w(g_warning ("Missing closing '>' bracket for email address: %s", *in));
-	}
-	
-	if (!name || !name->len) {
-		/* look for a trailing comment to use as the name? */
-		char *comment;
-		
-		if (name) {
-			g_string_free (name, TRUE);
-			name = NULL;
-		}
-		
-		comment = (char *) inptr;
-		decode_lwsp (&inptr);
-		if (inptr > comment) {
-			if ((comment = memchr (comment, '(', inptr - comment))) {
-				const char *cend;
-				
-				/* find the end of the comment */
-				cend = inptr - 1;
-				while (cend > comment && is_lwsp (*cend))
-					cend--;
-				
-				if (*cend == ')')
-					cend--;
-				
-				comment = g_strndup (comment + 1, (size_t) (cend - comment));
-				g_strstrip (comment);
-				
-				name = g_string_new (comment);
-				g_free (comment);
-			}
 		}
+	} else if (got_local) {
+		w(g_warning ("Missing '@' and domain in addr-spec: %.*s",
+			     inptr - start, start));
 	}
 	
 	*in = inptr;
 	
-	if (addr->len) {
-		if (name && !g_utf8_validate (name->str, name->len, NULL)) {
-			/* A (broken) mailer has sent us raw 8bit/multibyte text data... */
-			char *utf8 = g_mime_utils_decode_8bit (name->str, name->len);
-			
-			g_string_truncate (name, 0);
-			g_string_append (name, utf8);
-			g_free (utf8);
-		}
-		
-		mailbox = g_object_newv (INTERNET_ADDRESS_TYPE_MAILBOX, 0, NULL);
-		((InternetAddressMailbox *) mailbox)->addr = addr->str;
-		
-		if (name)
-			_internet_address_set_name (mailbox, name->str);
+	if (!got_local) {
+		w(g_warning ("Invalid addr-spec, missing local-part: %.*s",
+			     inptr - start, start));
+		g_string_free (addr, TRUE);
+		return NULL;
 	}
 	
-	g_string_free (addr, mailbox ? FALSE : TRUE);
-	if (name)
-		g_string_free (name, TRUE);
+	mailbox = g_object_newv (INTERNET_ADDRESS_TYPE_MAILBOX, 0, NULL);
+	((InternetAddressMailbox *) mailbox)->addr = addr->str;
+	g_string_free (addr, FALSE);
 	
 	return mailbox;
 }
 
 static InternetAddress *
-decode_address (const char **in)
+decode_group (const char **in)
 {
 	InternetAddressGroup *group;
+	InternetAddress *addr;
+	const char *inptr;
+	
+	inptr = *in;
+	
+	addr = internet_address_group_new (NULL);
+	group = (InternetAddressGroup *) addr;
+	
+	decode_lwsp (&inptr);
+	while (*inptr && *inptr != ';') {
+		InternetAddress *member;
+		
+		if ((member = decode_address (&inptr)))
+			_internet_address_group_add_member (group, member);
+		
+		decode_lwsp (&inptr);
+		while (*inptr == ',') {
+			inptr++;
+			decode_lwsp (&inptr);
+			if ((member = decode_address (&inptr)))
+				_internet_address_group_add_member (group, member);
+			
+			decode_lwsp (&inptr);
+		}
+	}
+	
+	*in = inptr;
+	
+	return addr;
+}
+
+static InternetAddress *
+decode_address (const char **in)
+{
+	const char *inptr, *start, *word, *comment = NULL;
 	InternetAddress *addr = NULL;
-	const char *inptr, *start;
-	const char *word;
+	gboolean has_lwsp = FALSE;
+	gboolean is_word;
 	GString *name;
 	
 	decode_lwsp (in);
 	start = inptr = *in;
 	
-	/* pre-scan */
 	name = g_string_new ("");
-	word = decode_word (&inptr);
 	
-	while (word) {
-		g_string_append_len (name, word, (size_t) (inptr - word));
-		if ((word = decode_word (&inptr)))
-			g_string_append_c (name, ' ');
-	}
-	
-	decode_lwsp (&inptr);
-	if (*inptr == ':') {
-		addr = internet_address_group_new (name->str);
-		group = (InternetAddressGroup *) addr;
-		inptr++;
-		
-		decode_lwsp (&inptr);
-		while (*inptr && *inptr != ';') {
-			InternetAddress *member;
+	/* Both groups and mailboxes can begin with a phrase (denoting
+	 * the display name for the address). Collect all of the
+	 * tokens that make up this name phrase.
+	 */
+	while (*inptr) {
+		if ((word = decode_word (&inptr))) {
+			g_string_append_len (name, word, (size_t) (inptr - word));
 			
-			if ((member = decode_mailbox (&inptr)))
-				_internet_address_group_add_member (group, member);
+		check_lwsp:
+			word = inptr;
+			skip_lwsp (&inptr);
 			
-			decode_lwsp (&inptr);
-			while (*inptr == ',') {
-				inptr++;
-				decode_lwsp (&inptr);
-				if ((member = decode_mailbox (&inptr)))
-					_internet_address_group_add_member (group, member);
-				
-				decode_lwsp (&inptr);
+			/* is the next token a word token? */
+			is_word = *inptr == '"' || is_atom (*inptr);
+			
+			if (inptr > word && is_word) {
+				g_string_append_c (name, ' ');
+				has_lwsp = TRUE;
 			}
+			
+			if (is_word)
+				continue;
 		}
 		
-		if (*inptr == ';')
+		/* specials    =  "(" / ")" / "<" / ">" / "@"  ; Must be in quoted-
+		 *             /  "," / ";" / ":" / "\" / <">  ;  string, to use
+		 *             /  "." / "[" / "]"              ;  within a word.
+		 */
+		if (*inptr == ':') {
+			/* group */
 			inptr++;
-		else
-			w(g_warning ("Invalid group spec, missing closing ';': %.*s",
-				     inptr - start, start));
-		
-		*in = inptr;
-	} else {
-		/* this is a mailbox */
-		addr = decode_mailbox (in);
+			addr = decode_group (&inptr);
+			decode_lwsp (&inptr);
+			if (*inptr != ';')
+				w(g_warning ("Invalid group spec, missing closing ';': %.*s",
+					     inptr - start, start));
+			else
+				inptr++;
+			break;
+		} else if (*inptr == '<') {
+			/* mailbox route-addr */
+			inptr++;
+			addr = decode_addrspec (&inptr);
+			decode_lwsp (&inptr);
+			if (*inptr != '>') {
+				w(g_warning ("Invalid route-addr, missing closing '>': %.*s",
+					     inptr - start, start));
+				
+				while (*inptr && *inptr != '>' && *inptr != ',')
+					inptr++;
+				
+				if (*inptr == '>')
+					inptr++;
+			} else
+				inptr++;
+			
+			/* if comment is non-NULL, we can check for a comment containing a name */
+			comment = inptr;
+			break;
+		} else if (*inptr == '(') {
+			/* beginning of a comment, use decode_lwsp() to skip past it */
+			decode_lwsp (&inptr);
+		} else if (*inptr && strchr ("@,;", *inptr)) {
+			if (name->len == 0) {
+				if (*inptr == '@') {
+					GString *domain;
+					
+					w(g_warning ("Unexpected address: %s: skipping.", start));
+					
+					/* skip over @domain? */
+					inptr++;
+					domain = g_string_new ("");
+					decode_domain (&inptr, domain);
+					g_string_free (domain, TRUE);
+				} else {
+					/* empty address */
+				}
+				break;
+			} else if (has_lwsp) {
+				/* assume this is just an unquoted special that we should
+				   treat as part of the name */
+				w(g_warning ("Unquoted '%c' in address name: %s: ignoring.", *inptr, start));
+				g_string_append_c (name, *inptr);
+				inptr++;
+				
+				goto check_lwsp;
+			}
+			
+		addrspec:
+			/* what we thought was a name was actually an addrspec? */
+			g_string_truncate (name, 0);
+			inptr = start;
+			
+			addr = decode_addrspec (&inptr);
+			
+			/* if comment is non-NULL, we can check for a comment containing a name */
+			comment = inptr;
+			break;
+		} else if (*inptr == '.') {
+			/* This would normally signify that we are
+			 * decoding the local-part of an addr-spec,
+			 * but sadly, it is common for broken mailers
+			 * to forget to quote/encode .'s in the name
+			 * phrase. */
+			g_string_append_c (name, *inptr);
+			inptr++;
+			
+			goto check_lwsp;
+		} else if (*inptr) {
+			/* Technically, these are all invalid tokens
+			 * but in the interest of being liberal in
+			 * what we accept, we'll ignore them. */
+			w(g_warning ("Unexpected char '%c' in address: %s: ignoring.", *inptr, start));
+			g_string_append_c (name, *inptr);
+			inptr++;
+			
+			goto check_lwsp;
+		} else {
+			goto addrspec;
+		}
 	}
 	
+	/* Note: will also skip over any comments */
+	decode_lwsp (&inptr);
+	
+	if (name->len == 0 && comment && inptr > comment) {
+		/* missing a name, look for a trailing comment */
+		if ((comment = memchr (comment, '(', inptr - comment))) {
+			const char *cend;
+			
+			/* find the end of the comment */
+			cend = inptr - 1;
+			while (cend > comment && is_lwsp (*cend))
+				cend--;
+			
+			if (*cend == ')')
+				cend--;
+			
+			g_string_append_len (name, comment + 1, (size_t) (cend - comment));
+		}
+	}
+	
+	if (addr && name->len > 0)
+		_internet_address_decode_name (addr, name);
+	
 	g_string_free (name, TRUE);
 	
+	*in = inptr;
+	
 	return addr;
 }
 
@@ -1619,9 +1571,9 @@ internet_address_list_parse_string (const char *str)
 		}
 		
 		decode_lwsp (&inptr);
-		if (*inptr == ',')
+		if (*inptr == ',') {
 			inptr++;
-		else if (*inptr) {
+		} else if (*inptr) {
 			w(g_warning ("Parse error at '%s': expected ','", inptr));
 			/* try skipping to the next address */
 			if ((inptr = strchr (inptr, ',')))
@@ -1636,31 +1588,3 @@ internet_address_list_parse_string (const char *str)
 	
 	return addrlist;
 }
-
-
-void
-_internet_address_list_block_event_handler (InternetAddressList *list, EventCallback callback, gpointer user_data)
-{
-	event_list_block (list->priv, callback, user_data);
-}
-
-
-void
-_internet_address_list_unblock_event_handler (InternetAddressList *list, EventCallback callback, gpointer user_data)
-{
-	event_list_unblock (list->priv, callback, user_data);
-}
-
-
-void
-_internet_address_list_add_event_handler (InternetAddressList *list, EventCallback callback, gpointer user_data)
-{
-	event_list_add (list->priv, callback, user_data);
-}
-
-
-void
-_internet_address_list_remove_event_handler (InternetAddressList *list, EventCallback callback, gpointer user_data)
-{
-	event_list_remove (list->priv, callback, user_data);
-}
